# this file contains code for our AST engine that parses given code, indexes it and then can be queried
# to get things. It's nothing as fancy as SCP, but it will need to have better UX than on_functions.
# Here's a list of usecases till now:
# - find a given object (str) in the given file and return all the relevant information about it

import re
import ast as A
from enum import Enum
from typing import List, Union

class IndexTypes(Enum):
  FUNCTION = A.FunctionDef
  CLASS = A.ClassDef
  VARIABLE = A.Name
  IMPORT = A.Import
  IMPORT_FROM = A.ImportFrom

  def all():
    return [IndexTypes.FUNCTION, IndexTypes.CLASS, IndexTypes.VARIABLE, IndexTypes.IMPORT, IndexTypes.IMPORT_FROM]


class RowItem:
  def __init__(self, name: str, type: IndexTypes, node: A.AST) -> None:
    self.name = name
    self.type = type
    self.node = node

  def __repr__(self) -> str:
    return f"{self.type.name} {self.name}"


class Visiboi(A.NodeVisitor):
  def __init__(self) -> None:
    self.data = []

  def visit_FunctionDef(self, node: A.FunctionDef) -> None:
    self.data.append(RowItem(node.name, IndexTypes.FUNCTION, node))

  def visit_ClassDef(self, node: A.ClassDef) -> None:
    self.data.append(RowItem(node.name, IndexTypes.CLASS, node))

  def visit_Name(self, node: A.Name) -> None:
    self.data.append(RowItem(node.id, IndexTypes.VARIABLE, node))

  def visit_Import(self, node: A.Import) -> None:
    for n in node.names:
      self.data.append(RowItem(n.name, IndexTypes.IMPORT, node))

  def visit_ImportFrom(self, node: A.ImportFrom) -> None:
    for n in node.names:
      self.data.append(RowItem(n.name, IndexTypes.IMPORT_FROM, node))


class Astea:
  def __init__(self, fname: str) -> None:
    """This is our AST engine, and it works on file level."""
    self.fname = fname
    with open(fname, 'r') as f:
      self.code = f.read()
    self.node = A.parse(self.code)
    self.vis = Visiboi()
    self.vis.visit(self.node)
    self.index = self.vis.data

  def find(self, x: str, types: Union[IndexTypes, List[IndexTypes]] = None) -> list:
    """Find all the instances of x in the file, and return a list of RowItems.
    
    - if there is a '.' (dot) in `x` find will try to perform a recursive search, you can escape the dot with a backslash
    """
    if types is None:
      types = IndexTypes.all()
    elif isinstance(types, IndexTypes):
      types = [types]

    # get the sections of the search string
    sections = []
    for s in re.split(r'(?<!\\)\.', x):
      sections.append(s.replace(r'\.', '.'))
    if len(sections) > 1:
      raise NotImplementedError("Recursive search not implemented yet")

    res = []
    for i, s in enumerate(sections):
      if i == 0:
        # first section
        res = [r for r in self.index if r.name == s and r.type in types]
      else:
        # recursive search
        res = [r for r in res if r.type == IndexTypes.CLASS]
        res = [r for r in res if r.node.body]
        res = [r for r in res if r.node.body[0].name == s]
    return res

    # if re.match(r'^\w+$', x):
    #   return [i for i in self.index if i.name == x and i.type in types]
    # elif re.match(r'^\w+\.\w+$', x):
    #   outer, inner = x.split('.')
    #   print(outer, inner)
    #   outer = self.find(outer, types)
    #   print(outer)
    #   if len(outer) == 0:
    #     return []
    #   outer = outer[0]
    #   for x in outer.node.body:
    #     print("-->>", x)
    #   return [i for i in self.index if i.name == inner and i.type in types and i.node in outer.node.body]
    # else:
    #   raise ValueError("Invalid search term")
